/*
A GLUT program using OpenGL, GLUI, and GLSL.

Demonstrates global illumination--computing incoming illumination.

Orion Sky Lawlor, olawlor@acm.org, 2007/04/12 (Public Domain)
*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#include <GL/glew.h> /* GL Extensions Wrangler, http://glew.sourceforge.net/ */
#include <GL/glut.h> /* GL Utilities Toolkit, http://freeglut.sourceforge.net/ */
#include <GL/glui.h> /* GL User Interface, http://glui.sourceforge.net/ */
#include "osl/vec4.h" /* GLSL-style C++ types, by Orion Lawlor */
#include "osl/mat4.h"
using osl::mat4;
#include "osl/osl.h" /* for osl::roundUp2 */

/* Dr. Lawlor's GLSL utilities */
#include "ogl/program.h"
#include "ogl/program.cpp"
#include "ogl/util.cpp"
#include "ogl/fast_mipmaps.c"


/* Variables updated by the GUI */
float distance=2.0; /* origin-to-camera distance */
int win_w, win_h; /**< Width and height (in pixels) of our window at last resize */

/* If true, regenerate lighting texture each frame */
int lighttex=1;
/* If true, show the lighting texture as it's computed */
int lighttex_verbose=0;
/* Number of pixels on a side in the lighting texture samples */
int lighttex_samps=16, lighttex_res=16;
/* Enable drawing various parts of the scene */
int draw_sky=1, draw_sphere=1, draw_torus=1, draw_light=1;
/* Parameters for rendering */
float brightness=2.0, lightsize=1.0, shininess=1.0;

/* These are rotation matrices, updated by the GLUI_Rotation spheres */
mat4 lightMat(1.0), modelMat(1.0), viewMat(1.0);

/// Scale by this offset
inline void glScalefv(const float *s) {
	glScalef(s[0],s[1],s[2]);
}
/// Translate by this offset
inline void glTranslatefv(const float *s) {
	glTranslatef(s[0],s[1],s[2]);
}
/** Transpose this mat4.  If the mat4 was a rotation matrix, 
  this actually inverts the matrix. */
mat4 transpose(const mat4 &m) {
	return mat4(
		m[0][0], m[1][0], m[2][0], m[3][0],
		m[0][1], m[1][1], m[2][1], m[3][1],
		m[0][2], m[1][2], m[2][2], m[3][2],
		m[0][3], m[1][3], m[2][3], m[3][3]);
}

#include <osl/matrixT.h>
/** Do a full-on inversion of this matrix */
mat4 invert_mat4(const mat4 &m) {
	osl::MatrixT<float,4,4> om;
	int r,c;
	for (r=0;r<4;r++) for (c=0;c<4;c++) om(r,c)=m[c][r];
	osl::MatrixT<float,4,4> im;
	if (!om.invert(im)) oglExit("matrix is not invertible!","invert_mat4");
	mat4 ret;
	for (r=0;r<4;r++) for (c=0;c<4;c++) ret[c][r]=im(r,c);
	return ret;
}


// These are the corners of our light source.
//  They get transformed by lightMat and lightsize.
vec3 lightCorners[4]={
	vec3(-1,2,-1), vec3(1,2,-1), vec3(1,2,1), vec3(-1,2,1)
	//vec3(-1,0,0), vec3(-1,-1,0), vec3(-1,-1,-1), vec3(-1,0,-1) 
};

/* Draw a big green ground polygon, and a big blue sky polygon */
void draw_ground(void) {
	double ground=100.0; /* extent of ground in X and Z */
	double groundY=-0.8;

// Ground
	glColor4f(0.5,0.7,0.6,1); /* green reflective */
	glMultiTexCoord3f(GL_TEXTURE3_ARB,0,0,0); /* no emission */
	glNormal3f(0,1,0); /* facing straight up */
	oglTextureQuad(
		oglVector3d(-ground,groundY,-ground), oglVector3d(+ground,groundY,-ground), 
		oglVector3d(-ground,groundY,+ground), oglVector3d(+ground,groundY,+ground)
	);

// Sky
	if (!draw_sky) return;
	groundY=10.0;
	glColor4f(0,0,0,1); /* no reflection */
	glMultiTexCoord3f(GL_TEXTURE3_ARB,0.2,0.3,0.9); /* blue emission */
	glNormal3f(0,-1,0); /* facing straight down */
	oglTextureQuad(
		oglVector3d(-ground,groundY,-ground), oglVector3d(+ground,groundY,-ground), 
		oglVector3d(-ground,groundY,+ground), oglVector3d(+ground,groundY,+ground)
	);
}

/* Draw the world */
void draw_world(void) 
{
	/* clear color and depth */
	glEnable(GL_SCISSOR_TEST);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	/* Silly hack: treat texture coordinate 3 as an emissive RGB color;
	   vertex color as RGB reflectance. */
	
	/* Draw a big ground polygon */
	draw_ground();
	
	// Draw the world 
	if (draw_torus) {
		glPushMatrix();
		glMultMatrixf(&modelMat[0][0]);
		glMultiTexCoord3f(GL_TEXTURE3_ARB,0,0,0); /* non-emissive */
		glColor4f(1,1,0,1); /* yellow */
		glutSolidTorus(0.1,0.7,4,6);
		glPopMatrix();
	}
	
	if (draw_sphere) {
		glMultiTexCoord3f(GL_TEXTURE3_ARB,1,0,0); /* sphere emits red light */
		glColor4f(1,0.2,0.2,1); /* reflects red too */
		glutSolidSphere(0.5,4,3);
	}
	
	if (draw_light) {
	// Draw the light source polygon
		glPushMatrix();
		glScalef(lightsize,lightsize,lightsize);
		glMultMatrixf(&lightMat[0][0]);
		glColor3f(0,0,0); /* light source reflects no light */
		glMultiTexCoord3f(GL_TEXTURE3_ARB,1,1,1); /* light source emits white light */
		glBegin(GL_TRIANGLE_FAN);
		for (int i=0;i<4;i++) glVertex3fv(lightCorners[i]);
		glEnd();
		glPopMatrix();
	}
	
}

/* Compute the incoming illumination at various locations and normals. 
  Internally, assumes a standard diffuse BRDF.
*/
class illumination_sampler {
public:
	illumination_sampler(int sz,oglProgramObject *p);
	
	/* Return the incoming illumination on a diffuse surface 
	   located at this point (P) with this normal (N). */
	vec3 sample(const vec3 &P,const vec3 &N);
private:
	int sz; oglProgramObject *p;
	typedef unsigned char pix_t;
	std::vector<pix_t> pix;
};

illumination_sampler::illumination_sampler(int sz_,oglProgramObject *p_)
	:sz(sz_), p(p_), pix(sz*sz*4)
{
	
}

/* Return the incoming illumination on a diffuse surface at this point */	
vec3 illumination_sampler::sample(const vec3 &P,const vec3 &N)
{
	/* Set up uniforms for cosine falloff */
	p->set("cameraloc",vec4(P,1.0));
	p->set("cameralook",vec4(normalize(N),0.0));
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	//Make a *wide* field-of-view camera:
	float fov=90.0; /* angular field-of-view, degrees */
	gluPerspective(fov, 1.0, 0.01,100.0);
	gluLookAt(P.x,P.y,P.z, P.x+N.x,P.y+N.y,P.z+N.z, 0.01,0.02,-1.0);
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	/* Draw the world */
	draw_world();
	
	/* Copy the rendered world back (slow!), and average */
	float r=0.0, g=0.0, b=0.0;
	glReadPixels(0,0,sz,sz, GL_RGBA,GL_UNSIGNED_BYTE, &pix[0]);
	for (int y=0;y<sz;y++)
	for (int x=0;x<sz;x++) {
		int i=4*(y*sz+x);
		r+=pix[i+0];
		g+=pix[i+1];
		b+=pix[i+2];
	}
	float normalize=1.0; /* value by which to scale entire view */
	float scl=normalize*(1.0/255)*(1.0/(sz*sz)); /* convert from summed-DN to average-float */
	return vec3(r*scl,g*scl,b*scl);
}

unsigned char to_byte(float f) {
	f*=255.0;
	if (f<0) return 0;
	if (f>255) return 255;
	else return (unsigned char)f;
}

/* GLUT calls this routine every time our window needs to be redrawn */
void main_display(void) {
	// Clear the color and depth buffers
	//glClearColor(0.2,0.4,0.8,0.0); /* sky blue */
	float g=0.0; glClearColor(g,g,g,0.0); /* gray */
	
	/* Depth testing */
	glEnable(GL_DEPTH_TEST);
	
	/* Alpha blending */
	glEnable(GL_BLEND);
	           /*  src color  + dest color  */
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	
	// Load all needed matrices into OpenGL
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity(); /*<- clean out any old leftover matrices */
	gluPerspective(90.0, /* <- Y camera field-of-view, in degrees */
		win_w/(float)win_h, /* viewport's aspect ratio: always width over height */
		0.01, /* Near clipping plane depth */
		100.0 /* Far clipping plane depth */
	);
	glTranslatef(0,0,-distance); /* push world origin down the *unrotated* Z axis */
	glMultMatrixf(&viewMat[0][0]); /* camera orientation: could also go under GL_MODELVIEW */
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
/* Set up programmable shader for lit objects */
	oglProgramObject *p=oglProgramFromFiles("vertex.txt","fragment.txt");
	p->begin();
	
	const int texX=osl::roundUp2(lighttex_res), texY=osl::roundUp2(lighttex_res), texZ=1; /* render this many total 3D samples */
	static unsigned int lighttex_handle=0u, lighttex_handle_size;
	if (lighttex_handle==0u || lighttex_handle_size!=texX) 
	{ /* first time around, initialize the 3D lighting texture */
		lighttex_handle_size=texX;
		glGenTextures(1,&lighttex_handle);
		GLenum target=GL_TEXTURE_3D;
		glBindTexture(target,lighttex_handle);
		glTexImage3D(target,0,GL_RGBA8,
			texX,texY,texZ,
			0,GL_RGBA,GL_UNSIGNED_BYTE,
			(void *)0);
		oglTextureType(oglTexture_linear,target);
		GLenum texWrap=GL_CLAMP_TO_EDGE; // GL_CLAMP_TO_BORDER; /* border color is zero, black */
		glTexParameteri(target, GL_TEXTURE_WRAP_S, texWrap); 
		glTexParameteri(target, GL_TEXTURE_WRAP_T, texWrap);	
		if (texZ>1) glTexParameteri(target, GL_TEXTURE_WRAP_R, texWrap);
		else 	glTexParameteri(target, GL_TEXTURE_WRAP_R, GL_REPEAT);	
	}
	
	glBindTexture(GL_TEXTURE_3D,lighttex_handle);
	p->set("lightTex",(int)0); /*<- texture unit activated */
	
	
	// Matrix mapping light tex coords (0,1) to world coords
	mat4 worldScale=mat4(1.0), swapTex=mat4(1.0);
	/* Swap Y and Z */
	swapTex[1][1]=swapTex[2][2]=0.0;
	swapTex[2][1]=swapTex[1][2]=1.0;
	/* Scale so X and Y coverages are reasonably big */
	float tx_s=6.0, ty_s=5.0; /* output span in X, Y */
	worldScale[0][0]*=tx_s;
	worldScale[3][0]+=-0.5*tx_s; /* center on origin */
	worldScale[1][1]*=ty_s;
	worldScale[3][1]+=-0.5*ty_s;
	worldScale[2][2]*=1.0;
	worldScale[3][2]+=-0.8; /* center outgoing Z on ground plane */
	mat4 discreteBias=mat4(1.0);
	discreteBias[3][0]-=0.5/texX;
	discreteBias[3][1]-=0.5/texY;
	discreteBias[3][2]-=0.5/texZ;
	mat4 worldFmTex=swapTex*worldScale*discreteBias;
	mat4 texFmWorld=invert_mat4(worldFmTex);
	p->set("lightTexFmWorld",texFmWorld);
	
	if (lighttex) 
	{ /* Render world from many points of view, for lighting */
		double start_time=oglTimer();
		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		
		int sz=lighttex_samps; /* render sz x sz pixels for each illumination sample */
		glViewport(0, 0, (GLsizei)sz, (GLsizei)sz);
		glScissor (0, 0, (GLsizei)sz, (GLsizei)sz);
		p->set("brightness",1.0);
		glBindTexture(GL_TEXTURE_3D,lighttex_handle);
		illumination_sampler samp(sz,p);
		
		// Direction illumination samples are facing
		vec3 N=normalize(vec3(0,1,0));
		mat4 worldFmPix=worldFmTex;
		worldFmPix[0]*=1.0/texX;
		worldFmPix[1]*=1.0/texY;
		worldFmPix[2]*=1.0/texZ;
		
		/* Compute illumination samples, and upload to 3D texture */
		unsigned char *lighttex_plane=new unsigned char[texX*texY*4];
		for (int z=0;z<texZ;z++) {
			for (int y=0;y<texY;y++)
			for (int x=0;x<texX;x++) {
				if (lighttex_verbose) {
					glViewport(x*sz, y*sz, (GLsizei)sz, (GLsizei)sz);
					glScissor (x*sz, y*sz, (GLsizei)sz, (GLsizei)sz);
				}
			
				/* Figure out where this texture pixel lands in world */
				vec3 P(x+0.5,y+0.5,z+0.5);
				P=vec3(worldFmPix*vec4(P,1.0));
				/* Grab illumination of world at that point */
				vec3 color=samp.sample(P,N);
				if (lighttex_verbose && x==(texX/2)) {
					printf("Color at tex(%d,%d,%d), world(%.3f,%.3f,%.3f): %f,%f,%f\n",
						x,y,z, 
						P.x,P.y,P.z,
						color.x,color.y,color.z);
				}
				int i=4*(y*texX+x);
				lighttex_plane[i+0]=to_byte(color.x);
				lighttex_plane[i+1]=to_byte(color.y);
				lighttex_plane[i+2]=to_byte(color.z);
				lighttex_plane[i+3]=0u;
			}
			glTexSubImage3D(GL_TEXTURE_3D,0,
				0,0,z,
				texX,texY,1,
				GL_RGBA,GL_UNSIGNED_BYTE,&lighttex_plane[0]);
		}
		delete[] lighttex_plane;
		glMatrixMode(GL_PROJECTION);
		glPopMatrix();
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		double elapsed=oglTimer()-start_time;
		printf("Rendering %d illumination samples (each %d x %d) took %.3f s\n",
			texX*texY*texZ, sz,sz, elapsed);
		if (lighttex_verbose) {glutSwapBuffers(); return;}
		
		//lighttex=0; /* don't re-light next frame (way too slow!) */
	}
	
	/* Render world from camera point of view */
	vec3 cameraloc=vec3(transpose(viewMat)*vec4(0,0,+distance,1));
	p->set("cameraloc",cameraloc);
	p->set("cameralook",normalize(-cameraloc)); /* camera is always looking at origin */
	p->set("brightness",brightness); /* full diffuse intensity */
	p->set("shininess",1.0); /* full specular intensity */
	glViewport(0, 0, (GLsizei)win_w, (GLsizei)win_h);
	glScissor (0, 0, (GLsizei)win_w, (GLsizei)win_h);
	
	draw_world();
	
	p->end();
	delete p;
	
	// Display the stuff drawn so far to the screen:
	glutSwapBuffers();
}

/* GLUT calls this routine every time our window changes size */
void main_reshape(int w, int h) {
	win_w=w; win_h=h;
	glViewport(0, 0, (GLsizei)win_w, (GLsizei)(win_h));
}

int main(int argc, char** argv)
{
	glutInitWindowSize(800,600);
	glutInit(&argc,argv);

	// Ask for a back buffer, color buffer, and Z buffer in a window.
	int mode=GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH;
	glutInitDisplayMode (mode);
	int main_window = glutCreateWindow("GLSL Radiosity Demo");

	// Turn on OpenGL extensions, like programmable shaders
	//   WARNING: you MUST call glutCreateWindow before doing this!
	//   If you don't call glewInit, or call it wrong, you'll segfault!
	glewInit();
	
	// Ask GLUT to call us when we need to redraw
	glutDisplayFunc (main_display);
	glutReshapeFunc (main_reshape);
	
	/* Set up view matrix so it's tilted forward a bit */
	vec3 up=normalize(vec3(0,1,0.2));
	viewMat[1]=vec4(up,0.0);
	viewMat[2]=vec4(vec3(up.x,-up.z,up.y),0.0);

	if (1) {
		// Make some sliders (this is the only GLUI code you'll need!)
		GLUI *glui = GLUI_Master.create_glui_subwindow(main_window, 
			GLUI_SUBWINDOW_TOP);
		/* This causes the main window to be redrawn when a GLUI control changes */
		glui->set_main_gfx_window( main_window );

		(new GLUI_Spinner( glui, "Distance:", &distance))
			->set_float_limits(0.0, 10.0);
		(new GLUI_Spinner( glui, "Lightsize:", &lightsize))
			->set_float_limits(0.0, 10.0);
		(new GLUI_Spinner( glui, "Shutter Speed:", &brightness))
			->set_float_limits(0.0, 100.0);
		
		new GLUI_Column(glui);
		new GLUI_Rotation(glui,"Light",&lightMat[0][0]);
		new GLUI_Column(glui);
		new GLUI_Rotation(glui,"Model",&modelMat[0][0]);
		new GLUI_Column(glui,false);
		new GLUI_Rotation(glui,"View",&viewMat[0][0]);
		new GLUI_Column(glui);
		new GLUI_Checkbox( glui, "Sky", &draw_sky);
		new GLUI_Checkbox( glui, "Light", &draw_light);
		new GLUI_Checkbox( glui, "Sphere", &draw_sphere);
		new GLUI_Checkbox( glui, "Torus", &draw_torus);
		new GLUI_Column(glui);
		new GLUI_Checkbox( glui, "Recompute Lighttex", &lighttex);
		new GLUI_Checkbox( glui, "Verbose Recompute", &lighttex_verbose);
		(new GLUI_Spinner( glui, "Sample Pixels:", &lighttex_samps))
			->set_int_limits(1, 100);
		(new GLUI_Spinner( glui, "Texture Size:", &lighttex_res))
			->set_int_limits(1, 512);
		
	}
	
	// Process events until user closes our window...
	glutMainLoop();
	
	return 0;
}
